#include <stdio.h>
#include <stdlib.h>
#include "mkdat.h"

//{{{
void init_table(void **table, int *cur_size, int new_size, int entry_size)
{
	if(*table != NULL || *cur_size != 0)
		return;
	
	*cur_size = new_size;
	*table = malloc(*cur_size * entry_size);
	
	if(*table == NULL)
		error_internal_memory();

}
//}}}
//{{{
void expand_table(void **table, int *cur_size, int new_size, int entry_size)
{
	if(new_size <= *cur_size)
		return;
	
	*cur_size *= 2;
	*table = realloc(*table, *cur_size * entry_size);
	
	if(*table == NULL)
		error_internal_memory();
}
//}}}

//{{{
// Input (str) starts with the character after a backslash, and continues at
// least to the end of the escape sequence. Advances the read ptr according to
// the length of the escape sequence.
char parse_escape(const char **str)
{
	unsigned char ret = 0;
	
	// Escape sequences are as per K&R, with the addition that non-special
	// characters are taken unchanged.
	switch(*((*str)++))
	{
		case 'a': // alert (bell)
			return '\a';
		case 'b': // backspace
			return '\b';
		case 'f': // formfeed
			return '\f';
		case 'n': // newline
			return '\n';
		case 'r': // carriage return
			return '\r';
		case 't': // horizontal tab
			return '\t';
		case 'v': // vertical tab
			return '\v';
		case '\\': // backslash
			return '\\';
		case '?': // question mark
			return '\?';
		case '\'': // single quote
			return '\'';
		case '\"': // double quote
			return '\"';
		case 'x': // hex number
			if(**str == '\0') {
				warn_escape_sequence();
				return '?';
			}
			if(**str >= 'a' && **str <= 'f')
				ret = (**str - 'a') << 4;
			else if(**str >= 'A' && **str <= 'F')
				ret = (**str - 'A') << 4;
			else if(**str >= '0' && **str <= '9')
				ret = (**str - '0') << 4;
			else {
				warn_escape_sequence();
				return '?';
			}
			(*str)++;
			if(**str >= 'a' && **str <= 'f')
				ret |= (**str - 'a');
			else if(**str >= 'A' && **str <= 'F')
				ret |= (**str - 'A');
			else if(**str >= '0' && **str <= '9')
				ret |= (**str - '0');
			else {
				warn_escape_sequence();
				return '?';
			}
			return ret;
		case '0': case '1': case '2': case '3': // octal number
			(*str)--;
			ret = (**str - '0') << 6;
			(*str)++;
			if(**str < '0' || **str > '7') {
				warn_escape_sequence();
				return '?';
			}
			ret |= (**str - '0') << 3;
			(*str)++;
			if(**str < '0' || **str > '7') {
				warn_escape_sequence();
				return '?';
			}
			ret |= **str - '0';
			(*str)++;
			return ret;
		case '\0': // Whoops! Past the end of the string!
			warn_escape_sequence();
			(*str)--;
			return '\\';
	}
	return **str;
}
//}}}
//{{{
int parse_char_constant(const char *str)
{
	str++;
	if(*str == '\\')
	{
		str++;
		return parse_escape(&str);
	}
	else
		return *str;
}
//}}}

//{{{
void numeric_bytes(char *data, int size)
{
	int i;
	char tmp;
	
	if(swap_endian)
	{
		for(i=0; i<size/2; i++)
		{
			tmp = data[i];
			data[i] = data[size-1-i];
			data[size-1-i] = tmp;
		}
	}

}
//}}}

//{{{
void error_internal_memory(void)
{
	fprintf(stderr, "Internal memory error!\n");
	exit(ERROR_INTERNAL);
}
//}}}
//{{{
void error_internal_param(void)
{
	fprintf(stderr, "Internal error: invalid parameter.\n");
	exit(ERROR_INTERNAL);
}
//}}}
//{{{
void error_undefined_identifier(const char *id)
{
	fprintf(stderr, "%s:%i: '%s' is undefined", output_filename, yylineno, id);
}
//}}}
//{{{
void warn_escape_sequence(void)
{
	fprintf(stderr, "%s:%i: warning: Invalid escape sequence", input_filename, yylineno);
}
//}}}
//{{{
void warn_range(int v)
{
	fprintf(stderr, "%s:%i: warning: Value %i outside range of data type\n", input_filename, yylineno, v);
}
//}}}
//{{{
void warn_dup_label(const char *identifier)
{
	fprintf(stderr, "%s:%i: warning: Duplicate label %s will be ignored\n", input_filename, yylineno, identifier);
}
//}}}
//{{{
void warn_undefined_reference(const char *identifier)
{
	fprintf(stderr, "%s:%i: warning: Undefined reference to '%s' converted to null pointer.\n", input_filename, yylineno, identifier);
}
//}}}
//{{{
void warn_multiple_definition(const char *identifier)
{
	fprintf(stderr, "%s:%i: warning: Multiple definition of '%s'.\n", input_filename, yylineno, identifier);
}
//}}}


